#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "mathlib.h"
#include "sph_cap.h"
#include "calcChar.h"
#include "ACHAIMDB.h"
#include "inGrid.h"
#include "ECHAIM_Chars.h"

//****DEBUGGING************
#ifdef DBG
#include "memwatch.h"
#endif
//*************************

//Generates the A-CHAIM characteristics
//Inputs:
//name = name of the desired characteristic (char *)
//lat = desires latitudes (geo-degrees) (double *)
//lon = desired longitudes (geo-degrees) (double *)
//nLat = length of lat and lon arrays (int)
//alt = desired altitudes (km)
//nAlt = legnth of altitude array (int)
//dbFile = path to the A-CHAIM output (char *)

//Possible name inputs (Case Sensitive):
//nmf2, hmf2, hbot, hf1, he, htop, hmf1

double *EC_Char (char *name, double *lat, double *lon, int nLat, char *dbFile)
{	
	//define variables
	static double *output;
	double	rotPar[9], *inside;
	sqlite3 *db; //sqlite3 database variable
	int rc, calcOpt=-1; //SQLite return code, calculation type
	
	//Open the DB file
	rc = sqlite3_open(dbFile, &db);
	if (rc != 0) printf("Error opening DB file (%i) %s\n", rc, dbFile);
	
	//get the calculation type
	calcOpt = getScalarDB(name, "TypeVal", &db);
	
	if (calcOpt == -1)
	{
		printf("Invalid characteristic name.\n");
		return 0;
	}
	
	//get the scalar data from the DB
	int nSphPar = getScalarDB(name, "NSph", &db);
	int nSphModelPar = getScalarDB(name, "NSphModel", &db);
	int capPar = getScalarDB(name, "Cap", &db);

	//Calculate the length of params and modelParams
	//Equation for sum of series sum((1..nSph)*2) + (nSph+1)
	int lp = ((nSphPar / 2.0) * ((2 * 2) + (nSphPar - 1) * 2)) + nSphPar + 1;
	int lmp = ((nSphModelPar / 2.0) * ((2 * 2) + (nSphModelPar - 1) * 2)) + nSphModelPar + 1;
	
	//get the vector data from the DB
	int tl; //temp length value
	double *tempArr = getVectorDB(name, "Rot", &tl, db);
	//check for expected length
	if (tl != 9) {printf("Error: Length of rotation matrix is unexpected.\n"); return 0;} 
	//transfer array to current scope
	for (int i=0; i<tl; i++) {rotPar[i] = tempArr[i];}
	free(tempArr);

	tempArr = getVectorDB(name, "Parameters", &tl, db);
	//check for expected length
	if (tl != lp) {printf("Error: Length of parameters matrix is unexpected.\n"); return 0;}
	//transfer array to current scope
	double params[tl];
	for (int i=0; i<tl; i++) {params[i] = tempArr[i];}
	free(tempArr);
	
	tempArr = getVectorDB(name, "ModelParameters", &tl, db);
	//check for expected length
	if (tl != lmp) {printf("Error: Length of model parameters matrix is unexpected.\n"); return 0;}
	//transfer array to current scope
	double modelParams[tl];
	for (int i=0; i<tl; i++) {modelParams[i] = tempArr[i];}
	free(tempArr);
	
	//get max value from nSph and nSphModel
	int nSph = (nSphPar > nSphModelPar) ? nSphPar : nSphModelPar;
	
	//calculate the basis sets for this function's characteristic using sph_cap
	//Use the basis sets to calculate the characteristic
	double **basis = sph_cap(lat, lon, nLat, nSph, capPar, rotPar);
	
	output = stat(basis, nLat, lmp, modelParams, lmp);
		
	//remove values outside the model grid
	inside = inGrid(rotPar, 45, lat, lon, nLat);
	for (int i=0; i<nLat; i++)
	{
		if (!inside[i]) {output[i] = 0.0 / 0.0;}
	}
		
	//free memory
	for (int i=0; i<nLat; i++) {free(basis[i]);}
	free(basis);
	
	//close the DB
	sqlite3_close(db);

	return output;
}

double *EC_Dens (double *lat, double *lon, double *alt, int nLat, char *dbFile)
{	
	//define variables
	static double *output;
	double dhbot = 1;
	
	//secondary parameters
	int r = 20;
	double g = 0.18, hme = 102;
	double k, ht, z, dhf1, dhe, h, densBot, densTop;
	
	//alocate memory
	output = calloc(nLat, sizeof(double));
	
	//get the ionospheric characteristics
	double *nmf2 = EC_Char("nmf2", lat, lon, nLat, dbFile);
	double *hmf2 = EC_Char("hmf2", lat, lon, nLat, dbFile);
	double *hmf1 = EC_Char("hmf1", lat, lon, nLat, dbFile);
	double *hbot = EC_Char("hbot", lat, lon, nLat, dbFile);
	double *htop = EC_Char("htop", lat, lon, nLat, dbFile);
	double *hf1 = EC_Char("hf1", lat, lon, nLat, dbFile);
	double *he = EC_Char("he", lat, lon, nLat, dbFile);
	
	//Calculate the density
	for (int i=0; i<nLat; i++)
	{
		if (alt[i] >= hmf2[i])
		{
			k = alt[i] - hmf2[i];
			ht = htop[i] * (1 + ((r * g * k) / ((r * htop[i]) + (g * k))));
			z = k / ht;
			
			densTop = nmf2[i] * pow(sech(z / 2), 2);
			densBot = 0;
		} else
		{
			k = alt[i] - hmf2[i];
			dhf1 =  pow(sech((k) / (0.5 * (hmf2[i] - hmf1[i]))), 2);
			dhe = pow(sech((alt[i] - hme) / 30.0), 2);
			h = (hbot[i] * dhbot + hf1[i] * dhf1 + he[i] * dhe) * \
				(1.0 / (1.0 + exp((hme - 15.0 - alt[i]) / 2.5)));
				
			densBot = nmf2[i] * pow(sech(k / h), 2);
			densTop = 0;
		}
		
		output[i] = densTop + densBot;
	}
	
	return output;
}

double *EC_TEC (double *lat, double *lon, int nLat, char *dbFile)
{	
	//define variables
	static double *output;
	int nAlt = (((550 - 85) / 15)  + 1) + (((2050 - 650) / 100) + 1); //85 - 550 km at 15 km res, 550 - 2050 km at 100 km res
	double *dens, alt[nAlt], latIn[nAlt], lonIn[nAlt];
	
	//alocate memory
	output = calloc(nLat, sizeof(double));
	
	//set the alt array
	alt[0] = 85;
	for (int i = 1; i < nAlt; i++)
	{
		alt[i] = (alt[i-1] < 550) ? alt[i-1] + 15 : alt[i-1] + 100;
	}
	
	//calculate TEC for each lat/lon pair
	for (int i = 0; i < nLat; i++)
	{
		for (int j = 0; j < nAlt; j++)
		{
			latIn[j] = lat[i];
			lonIn[j] = lon[i];
		}
				
		//get the density measurements
		dens = EC_Dens(latIn, lonIn, alt, nAlt, dbFile);
		
		//Calculate TEC
		output[i] = simpsonsInt(alt, dens, nAlt) / (pow(10,13));
	}
	
	return output;
}