#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mathlib.h"
#include "sph_cap.h"

#ifndef M_PI
#    define M_PI 3.14159265358979323846
#endif

#define DEG2RAD 0.0174532925199432957692369076

//****DEBUGGING************
#ifdef DBG
#include "memwatch.h"
#endif
//*************************

//Calculate the spherical cap coordinates
//Pass an array of altitude and longitudes,
//The max of nSph/nSphModel, Cap value
//Rotation matrix (expects 9 element vector, translating to 3x3 matrix
double **sph_cap(double *lat, double *lon, int nLat, int pMax, int cap, double *rot)
{
	//define variables
	double *ones; //array of ones
	double *coLat, capRad; //co-latitudes, radian cap angle
	double *onesSph, *latSph, *lonSph, *latRad, *lonRad; //arrays for steps of transformed coordinates
	static double **output; //output array
	
	//allocate arrays
	ones = calloc(nLat, sizeof(double));
	latRad = calloc(nLat, sizeof(double));
	lonRad = calloc(nLat, sizeof(double));
	onesSph = calloc(nLat, sizeof(double));
	latSph = calloc(nLat, sizeof(double));
	lonSph = calloc(nLat, sizeof(double));
	coLat = calloc(nLat, sizeof(double));
	
	//calculate size of output array
	//Equation for sum of series sum((1..pMax)*2) + (pMax+1)
	int n = ((pMax / 2) * ((2 * 2) + (pMax - 1) * 2)) + pMax + 1;
	
	//allocate output array
	output = calloc(nLat, sizeof(double *));
	for (int i=0; i<nLat; i++) {output[i] = calloc(n, sizeof(double));}
	
	//convert latitude and longitude to radians
	//and populate the ones array
	for (int i=0; i<nLat; i++)
	{
		latRad[i] = lat[i] * DEG2RAD;
		lonRad[i] = lon[i] * DEG2RAD;
		ones[i] = 1;
	}
	capRad = (double)cap * DEG2RAD;
	
	//convert latitude/longitude to spherical
	cart2sph(ones, latRad, lonRad, nLat, -1, onesSph, latSph, lonSph);
	
	//rotate the spherical coordinates
	double tOnes = 0, tLat = 0, tLon = 0; // temp variables
	for (int i=0; i<nLat; i++)
	{
		tOnes = onesSph[i];
		tLat = latSph[i];
		tLon = lonSph[i];
		
		onesSph[i] = (tOnes * rot[0]) + (tLat * rot[1]) + (tLon * rot[2]);
		latSph[i] = (tOnes * rot[3]) + (tLat * rot[4]) + (tLon * rot[5]);
		lonSph[i] = (tOnes * rot[6]) + (tLat * rot[7]) + (tLon * rot[8]);
	}
	
	//Convert rotated coordinates to cartesian
	cart2sph(onesSph, latSph, lonSph, nLat, 1, ones, latRad, lonRad);
	
	//calculate co-latitude
	for (int i=0; i<nLat; i++)
	{
		coLat[i] = M_PI * ((M_PI / 2) - latRad[i]) / ((M_PI / 2) - capRad);
	}
	
	//Calculate the output
	int ind = 0;
	double pl = 0, schNorm = 0, mm = 0, nlm = 0;
	for (int n=0; n<=pMax; n++)
	{
		for (int m=0; m<=n; m++)
		{
			for (int c=0; c<nLat; c++)
			{
				//Get legendre polynomial
				pl = legendre(cos(coLat[c]), n, m);
				
				if (m != 0)
				{
					//Calculate Schmidt seminormalized factor
					schNorm = pow(-1,m) * sqrt((2 * fact(n - m)) / fact(n + m));
					pl *= schNorm;
				}
				
				//calculate NLM function
				mm = (m == 0) ? 1 : 0;
				nlm = pow(-1,m) * sqrt((2 - mm) * ((2 * n) + 1));
				pl *= nlm;
				
				output[c][ind] = pl * cos(m * lonRad[c]);
				
				if (m != 0)
				{
					output[c][ind+1] = pl * sin(m * lonRad[c]);
					c++;c--;
				}
			}
			
			ind = (m == 0) ? ind + 1 : ind + 2;
		}
		
	}
	
	//free memory
	free(coLat);
	free(latSph);
	free(lonSph);
	free(ones);
	free(onesSph);
	free(lonRad);
	free(latRad);
	
	return output;
}


//Convert between spherical and Cartesian coordinates
//Inputs:
//Input coordinates as vector (one for each coordinate) (double *)
//Direction of conversion (positive = cart 2 sph, negative = sph 2 cart)
//Output:
//n element vector of converted input (each coordinate) (double *)
void cart2sph(double *inX, double *inY, double *inZ, int nIn, int dir, \
				double *outX, double *outY, double *outZ)
{	
	if (dir == 0)
	{	
		printf("Direction input must be a positive or negative integer.\n");
		return;
	}
	
	if (dir > 0)
	{
		for (int i=0; i<nIn; i++)
		{
			outX[i] = sqrt(pow(inX[i],2) + pow(inY[i],2) + pow(inZ[i],2));
			outY[i] = atan2(inZ[i], sqrt(pow(inX[i],2) + pow(inY[i],2)));
			outZ[i] = atan2(inY[i], inX[i]);
		}
		
		
	}else
	{
		for (int i=0; i<nIn; i++)
		{
			outX[i] = inX[i] * cos(inY[i]) * cos(inZ[i]);
			outY[i] = inX[i] * cos(inY[i]) * sin(inZ[i]);
			outZ[i] = inX[i] * sin(inY[i]);
		}
	}
}
